from django import forms
from django.contrib.auth import get_user_model
from django.core.mail import send_mail
from django.template import loader
from django.utils.module_loading import import_string
from magicauth import settings as magicauth_settings
from magicauth.models import MagicToken
import math

email_unknown_callback = import_string(magicauth_settings.EMAIL_UNKNOWN_CALLBACK)


class EmailForm(forms.Form):
    email = forms.EmailField()

    def clean_email(self):
        user_email = self.cleaned_data["email"]
        user_email = user_email.lower()

        email_field = magicauth_settings.EMAIL_FIELD
        field_lookup = {f"{email_field}__iexact": user_email}
        if not get_user_model().objects.filter(**field_lookup).exists():
            email_unknown_callback(user_email)
        return user_email

    def create_token(self, user):
        token = MagicToken.objects.create(user=user)
        return token

    def send_email(self, current_site, next_url):
        user_email = self.cleaned_data["email"]
        email_field = magicauth_settings.EMAIL_FIELD
        field_lookup = {f"{email_field}__iexact": user_email}
        user = get_user_model().objects.get(**field_lookup)
        token = self.create_token(user)
        email_subject = magicauth_settings.EMAIL_SUBJECT
        html_template = magicauth_settings.EMAIL_HTML_TEMPLATE
        text_template = magicauth_settings.EMAIL_TEXT_TEMPLATE
        from_email = magicauth_settings.FROM_EMAIL
        context = {
            "token": token,
            "user": user,
            "site": current_site,
            "next_url": next_url,
            "TOKEN_DURATION_MINUTES": math.floor(magicauth_settings.TOKEN_DURATION_SECONDS / 60),
            "TOKEN_DURATION_SECONDS": magicauth_settings.TOKEN_DURATION_SECONDS,
        }
        text_message = loader.render_to_string(text_template, context)
        html_message = loader.render_to_string(html_template, context)
        send_mail(
            subject=email_subject,
            message=text_message,
            from_email=from_email,
            html_message=html_message,
            recipient_list=[user_email],
            fail_silently=False,
        )



from django.contrib.auth import get_user_model

import factory
from faker import Factory as FakerFactory

from pytest_factoryboy import register

faker = FakerFactory.create("fr_FR")


@register
class UserFactory(factory.DjangoModelFactory):
    email = factory.LazyFunction(faker.email)
    username = factory.LazyAttribute(lambda a: a.email)

    class Meta:
        model = get_user_model()


@register
class MagicTokenFactory(factory.DjangoModelFactory):
    user = factory.SubFactory(UserFactory)

    class Meta:
        model = "magicauth.MagicToken"



from datetime import timedelta
from pytest import mark
import urllib.parse

from django.core import mail
from django.shortcuts import reverse
from django.utils import timezone
from magicauth.models import MagicToken
from tests import factories

pytestmark = mark.django_db


def test_getting_LoginView_while_authenticated_redirects_to_default(client):
    user = factories.UserFactory()
    client.force_login(user)
    url = reverse("magicauth-login")
    response = client.get(url)
    assert response.status_code == 302
    assert response.url == "/landing/"


def test_getting_LoginView_while_authenticated_with_next_redirects_to_next(client):
    user = factories.UserFactory()
    client.force_login(user)
    url = reverse("magicauth-login") + "?next=/test_dashboard/"
    response = client.get(url)
    assert response.status_code == 302
    assert response.url == "/test_dashboard/"


def test_posting_email_for_valid_existing_user_redirects(client):
    user = factories.UserFactory()
    url = reverse("magicauth-login")
    data = {"email": user.email}
    response = client.post(url, data=data)
    assert response.status_code == 302
    assert len(mail.outbox) == 1


def test_loging_with_email_is_case_insensitive(client):
    user = factories.UserFactory()
    url = reverse("magicauth-login")
    data = {"email": user.email.upper()}
    response = client.post(url, data=data)
    assert response.status_code == 302
    assert len(mail.outbox) == 1


def test_posting_unknown_email_raise_error_and_dont_send_email(client):
    url = reverse("magicauth-login")
    data = {"email": "unknown@email.com"}
    response = client.post(url, data=data)
    assert "invalid" in str(response.content)
    assert len(mail.outbox) == 0


def test_authenticated_user_is_redirected_to_default_redirect_page(client):
    user = factories.UserFactory()
    client.force_login(user)
    url = reverse("magicauth-login")
    response = client.get(url)
    assert response.status_code == 302
    assert response.url == "/landing/"


def test_authenticated_user_is_redirected_to_next_url(client):
    user = factories.UserFactory()
    client.force_login(user)
    url = reverse("magicauth-login")
    response = client.get(url, {"next": "/test_dashboard/"})
    assert response.status_code == 302
    assert response.url == "/test_dashboard/"


def test_posting_email_for_valid_existing_user_sends_email(client):
    user = factories.UserFactory()
    url = reverse("magicauth-login")
    data = {"email": user.email}
    client.post(url, data=data)
    assert len(mail.outbox) == 1


def test_posting_email_redirect_to_default_view(client):
    user = factories.UserFactory()
    url = reverse("magicauth-login")
    data = {"email": user.email}
    client.post(url, data=data)
    assert "?next=/landing/" in mail.outbox[0].body


def test_posting_email_with_next_redirects_to_next(client):
    user = factories.UserFactory()
    url = reverse("magicauth-login") + "?next=/test_dashboard/"
    data = {"email": user.email}
    client.post(url, data=data)
    assert "?next=/test_dashboard/" in mail.outbox[0].body


def test_posting_unknown_email_does_not_send_email(client):
    url = reverse("magicauth-login")
    data = {"email": "unknown@email.com"}
    client.post(url, data=data)
    assert len(mail.outbox) == 0


def test_posting_email_for_valid_existing_user_created_token(client):
    user = factories.UserFactory()
    url = reverse("magicauth-login")
    data = {"email": user.email}
    count_before = MagicToken.objects.count()
    client.post(url, data=data)
    count_after = MagicToken.objects.count()
    assert count_after == count_before + 1


def test_opening_magic_link_with_valid_token_redirects(client):
    token = factories.MagicTokenFactory()
    url = reverse("magicauth-validate-token", args=[token.key])
    response = client.get(url)
    assert response.status_code == 302
    assert response.url == "/landing/"


def test_opening_magic_link_with_a_next_sets_a_new_url(client):
    token = factories.MagicTokenFactory()
    next_url_raw = "/test_dashboard/?a=test&b=test"
    # We use `quote` because the URL has parameters:
    next_url = urllib.parse.quote(next_url_raw)
    validate_token_url = reverse("magicauth-validate-token", kwargs={"key": token.key})
    url = f'{validate_token_url}?next={next_url}'
    response = client.get(url)
    assert response.status_code == 302
    assert response.url == next_url_raw


def test_login_page_raises_404_if_unsafe_next_url(client):
    token = factories.MagicTokenFactory()
    url = (
        reverse("magicauth-validate-token", kwargs={"key": token.key})
        + "?next=http://www.myfishingsite.com/?a=test&b=test"
    )
    response = client.get(url)
    assert response.status_code == 404


def test_validate_token_view_raises_404_if_unsafe_next_url(client):
    token = factories.MagicTokenFactory()
    url = (
        reverse("magicauth-validate-token", kwargs={"key": token.key})
        + "?next=http://www.myfishingsite.com/?a=test&b=test"
    )
    response = client.get(url)
    assert response.status_code == 404


def test_validate_token_view_raises_404_for_loggedin_user_if_unsafe_next_url(client):
    token = factories.MagicTokenFactory()
    user = factories.UserFactory()
    client.force_login(user)
    url = (
        reverse("magicauth-validate-token", kwargs={"key": token.key})
        + "?next=http://www.myfishingsite.com/?a=test&b=test"
    )
    response = client.get(url)
    assert response.status_code == 404
    assert user.is_authenticated


def test_wait_page_raises_404_if_unsafe_next_url(client):
    token = factories.MagicTokenFactory()
    url = (
        reverse("magicauth-wait", kwargs={"key": token.key})
        + "?next=http://www.myfishingsite.com/?a=test&b=test"
    )
    response = client.get(url)
    assert response.status_code == 404


def test_email_sent_page_raises_404_if_unsafe_next_url(client):
    url = (
        reverse("magicauth-email-sent")
        + "?next=http://www.myfishingsite.com/?a=test&b=test"
    )
    response = client.get(url)
    assert response.status_code == 404


def test_token_is_removed_after_visiting_magic_link(client):
    token = factories.MagicTokenFactory()
    url = reverse("magicauth-validate-token", args=[token.key])
    count_before = MagicToken.objects.count()
    client.get(url)
    count_after = MagicToken.objects.count()
    assert count_after == count_before - 1


def test_duplicate_token_for_same_user_is_removed_after_visiting_magic_link(client):
    token = factories.MagicTokenFactory()
    duplicate = factories.MagicTokenFactory(user=token.user)
    url = reverse("magicauth-validate-token", args=[token.key])
    client.get(url)
    assert duplicate not in MagicToken.objects.all()


def test_visiting_magic_link_triggers_login(client):
    token = factories.MagicTokenFactory()
    url = reverse("magicauth-validate-token", args=[token.key])
    client.get(url)
    assert "_auth_user_id" in client.session


def test_unknown_token_redirects(client):
    url = reverse("magicauth-validate-token", args=["unknown-token"])
    response = client.get(url)
    assert response.status_code == 302


def test_expired_token_redirects(client):
    token = factories.MagicTokenFactory()
    token.created = timezone.now() - timedelta(days=1)
    token.save()
    url = reverse("magicauth-validate-token", args=[token.key])
    response = client.get(url)
    assert response.status_code == 302


def test_expired_token_is_deleted(client):
    token = factories.MagicTokenFactory()
    token.created = timezone.now() - timedelta(days=1)
    token.save()
    url = reverse("magicauth-validate-token", args=[token.key])
    client.get(url)
    assert token not in MagicToken.objects.all()



from django.urls import path

from . import views as magicauth_views
from . import settings as magicauth_settings

urlpatterns = [
    path(
        magicauth_settings.LOGIN_URL,
        magicauth_views.LoginView.as_view(),
        name="magicauth-login",
    ),
    path(
        magicauth_settings.EMAIL_SENT_URL,
        magicauth_views.EmailSentView.as_view(),
        name="magicauth-email-sent",
    ),
    path(
        magicauth_settings.WAIT_URL,
        magicauth_views.WaitView.as_view(),
        name="magicauth-wait",
    ),
    path(
        magicauth_settings.VALIDATE_TOKEN_URL,
        magicauth_views.ValidateTokenView.as_view(),
        name="magicauth-validate-token",
    ),
]



from django.urls import path, include
from django.views.generic import TemplateView

urlpatterns = [
    path("", include("magicauth.urls")),
    path(
        "landing/", TemplateView.as_view(template_name="home.html"), name="test_home"
    ),
]



import os
from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), "README.md")) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name="django-magicauth",
    version="0.4",
    packages=find_packages(),
    include_package_data=True,
    license="",
    description=(
        "Password-less authentication: login by clicking on "
        "a magic link received by email."),
    long_description=README,
    long_description_content_type='text/markdown',
    url="",
    author="",
    author_email="",
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        'License :: OSI Approved :: BSD License',
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
)



from datetime import timedelta
import logging
import urllib.parse

from django.contrib import messages
from django.contrib.auth import login
from django.http import Http404
from django.shortcuts import redirect
from django.urls import reverse_lazy, reverse
from django.utils import timezone
from django.utils.http import is_safe_url
from django.views.generic import View, FormView, TemplateView

from magicauth import settings as magicauth_settings
from magicauth.forms import EmailForm
from magicauth.models import MagicToken

logger = logging.getLogger()


class NextUrlMixin(object):
    """
    Helper for managing the 'next url' parameter.
    """

    def get_next_url(self, request):
        """
        Get the next url from the querystring parameters (?next=/my/next/page).
        If the next parameter is not there, returns the default redirect url
        """
        next_url = request.GET.get("next")
        if not next_url:
            next_url = reverse(magicauth_settings.LOGGED_IN_REDIRECT_URL_NAME)
        # the following `is_safe_url` will be deprecated in django 4 and replaced by
        # url_has_allowed_host_and_scheme
        if not is_safe_url(next_url, allowed_hosts={request.get_host()}, require_https=True):
            # We are not logging the unsafe URL to prevent code injections in logs
            logger.warning("[MagicAuth] an unsafe URL was used through a login link")
            raise Http404
        return next_url

    def get_next_url_encoded(self, request):
        """
        Use this when the URL needs to be encoded, for instance when including the URL to string
        before a redirect.
        """
        url = self.get_next_url(self.request)
        return urllib.parse.quote(url)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        next_url = self.get_next_url(self.request)
        context["next_url"] = next_url
        return context


class LoginView(NextUrlMixin, FormView):
    """
    The login page. The user enters their email in the form to get a link by email.
    """

    form_class = EmailForm
    success_url = reverse_lazy("magicauth-email-sent")
    template_name = magicauth_settings.LOGIN_VIEW_TEMPLATE

    def get(self, request, *args, **kwargs):
        if request.user.is_authenticated:
            next_url = self.get_next_url(self.request)
            return redirect(next_url)
        return super(LoginView, self).get(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super(LoginView, self).get_context_data(**kwargs)
        context[
            "LOGGED_IN_REDIRECT_URL_NAME"
        ] = magicauth_settings.LOGGED_IN_REDIRECT_URL_NAME
        context["LOGOUT_URL_NAME"] = magicauth_settings.LOGOUT_URL_NAME
        return context

    def get_success_url(self, **kwargs):
        url = reverse_lazy("magicauth-email-sent")
        # Use encoded next URL before including it in a string
        next_url_quoted = self.get_next_url_encoded(self.request)
        return f"{url}?next={next_url_quoted}"

    def form_valid(self, form, *args, **kwargs):
        next_url = self.get_next_url(self.request)
        current_site = self.request.site
        form.send_email(current_site, next_url)
        return super().form_valid(form)


class EmailSentView(NextUrlMixin, TemplateView):
    """
    View shown to confirm the email has been sent.
    """
    template_name = magicauth_settings.EMAIL_SENT_VIEW_TEMPLATE


class WaitView(NextUrlMixin, TemplateView):
    """
    Wait few seconds before redirecting to login.
    This is for solving an issue with a security feature in some email clients where
    the magic link is verified and and thus the token gets invalidated.
    """
    template_name = magicauth_settings.WAIT_VIEW_TEMPLATE

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        token_key = kwargs.get("key")
        validate_token_url = reverse('magicauth-validate-token', kwargs={'key': token_key})
        context["validate_token_url"] = validate_token_url
        context["WAIT_SECONDS"] = magicauth_settings.WAIT_SECONDS
        return context


class ValidateTokenView(NextUrlMixin, View):
    """
    The link sent by email goes to this view.
    It validates the token passed in querystring,
    and either logs in or shows a form to make a new token.
    """

    @staticmethod
    def get_valid_token(key):
        duration = magicauth_settings.TOKEN_DURATION_SECONDS
        try:
            token = MagicToken.objects.get(key=key)
        except MagicToken.DoesNotExist:
            return None
        except MagicToken.MultipleObjectsReturned:
            return None

        if token.created < timezone.now() - timedelta(seconds=duration):
            token.delete()
            return None
        return token

    def get(self, request, *args, **kwargs):
        url = self.get_next_url(request)
        if request.user.is_authenticated:
            return redirect(url)
        token_key = kwargs.get("key")
        token = self.get_valid_token(token_key)
        if not token:
            messages.warning(
                self.request,
                "Ce lien de connexion ne fonctionne plus. "
                "Pour en recevoir un nouveau, nous vous invitons à renseigner "
                "votre email ci-dessous puis à cliquer sur valider.",
            )
            return redirect("magicauth-login")
        login(self.request, token.user)
        MagicToken.objects.filter(
            user=token.user
        ).delete()  # Remove them all for this user
        return redirect(url)
